from django.contrib import admin
from .models import VSServer, VSMod


class VSModsInline(admin.TabularInline):
    model = VSMod


class VSServerAdmin(admin.ModelAdmin):
    readonly_fields = ['last_heartbeat',]
    inlines = [VSModsInline]


admin.site.register(VSServer, VSServerAdmin)
admin.site.register(VSMod)



import json
from django.http import JsonResponse
from django.utils import timezone
from django.utils.decorators import method_decorator
from django.views.decorators.csrf import csrf_exempt
from django.views import View

import socket

from .utils import get_client_ip, vs_prune_old_servers
from .models import VSServer


class CSRFExemptMixin:
   @method_decorator(csrf_exempt)
   def dispatch(self, *args, **kwargs):
       return super(CSRFExemptMixin, self).dispatch(*args, **kwargs)


class VSMasterRegister(CSRFExemptMixin, View):
    def post(self, request):
        data = json.loads(request.body)
        ip = get_client_ip(request)

        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
            # Check if port is open -- probably a better way to do this?
            result = sock.connect_ex((ip, data['port']))
            if result == 0:
                server = VSServer.objects.create_server(data, ip)
                # vs_check_server_still_alive(server.token)
                resp = {"data": server.token, "status": "ok"}
            else:
                resp = {"status":"bad", "data":"port not opened"}

        return JsonResponse(resp)


class VSMasterHeartbeat(CSRFExemptMixin, View):
    def post(self, request):
        data = json.loads(request.body)
        server: VSServer = VSServer.objects.filter(token=data['token']).first()
        if server:
            server.last_heartbeat = timezone.now()
            server.players = data['players']
            server.save()
            return JsonResponse({"status": "ok"})
        else:
            return JsonResponse({"status": "timeout", "data": "timeout"})  # game server should re-register itself


class VSMasterUnregister(CSRFExemptMixin, View):
    def post(self, request):
        data = json.loads(request.body)
        server: VSServer = VSServer.objects.filter(token=data['token']).first()
        server.delete()
        return JsonResponse({"status": "ok"})


class VSMasterList(CSRFExemptMixin, View):
    def get(self, request):
        # got only servers that have provided a heartbeat recently.
        servers = VSServer.objects.filter(last_heartbeat__gte=timezone.now()-timezone.timedelta(minutes=6))
        server_list = [server.as_dict() for server in servers]
        return JsonResponse(
            {
                "status": "ok",
                "data": server_list
            }
        )

class VSMasterPruneDB(CSRFExemptMixin, View):
    """
    In theory servers should unregister themselves when they exit, and will be gracefully removed from the database.
    In reality someone is going to force kill a server and leave it hanging around. Ideally you will set up a background
    task using cron/celery/whatever that periodically wipes the database of aged entries by calling
    `utils.vs_prune_old_servers()`

    But, if you don't care about the potential issues, this example endpoint is provided (but not given a URL by
    default) to trigger such a pruning so you can visit the an appropriate url to cause a prune on-demand. You're likely
    better off writing your own view for this.
    """
    def get(self, request):
        vs_prune_old_servers(10)



# Generated by Django 5.1.6 on 2025-03-01 17:24

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='VSServer',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('vs_server_name', models.CharField(max_length=512)),
                ('address', models.CharField(max_length=256)),
                ('port', models.PositiveIntegerField()),
                ('play_style_id', models.CharField(max_length=64)),
                ('play_style_lang_code', models.CharField(max_length=64)),
                ('is_whitelisted', models.BooleanField()),
                ('description', models.TextField()),
                ('max_players', models.SmallIntegerField()),
                ('game_version', models.CharField(max_length=64)),
                ('has_password', models.BooleanField()),
                ('server_url', models.CharField(max_length=256, null=True)),
                ('last_heartbeat', models.DateTimeField(auto_now_add=True)),
                ('players', models.SmallIntegerField(default=0)),
                ('token', models.CharField(max_length=64)),
            ],
        ),
        migrations.CreateModel(
            name='VSMod',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('_id', models.CharField(max_length=64)),
                ('version', models.CharField(max_length=64)),
                ('server', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='mods', to='vintagestory_masterserver.vsserver')),
            ],
        ),
    ]



from django.urls import path

from . import views

app_name = "vsmaster"

urlpatterns = [
    path('register', views.VSMasterRegister.as_view(), name='vsregister'),
    path('heartbeat', views.VSMasterHeartbeat.as_view(), name="vsheartbeat"),
    path('unregister', views.VSMasterUnregister.as_view(), name="vsunregister"),
    path('list', views.VSMasterList.as_view(), name="vslist"),
]



import base64
import secrets
from typing import TypedDict, List
from django.db import models


class VSplaystyleData(TypedDict):
    id: str
    langCode: str


class VSmodData(TypedDict):
    id: str
    version: str


class VSServerData(TypedDict):
    port: int
    name: str
    icon: str
    playstyle: VSplaystyleData
    maxPlayers: int
    gameVersion: str
    hasPassword: bool
    Mods: List[VSmodData]
    serverUrl: str | None
    gameDescription: str
    whitelisted: bool


class VSServerManager(models.Manager):
    def create_server(self, data: VSServerData, ip_addr: str):
        server = self.create(
            vs_server_name = data['name'],
            address = ip_addr,
            port=data['port'],
            play_style_id=data['playstyle']['id'],
            play_style_lang_code=data['playstyle']['langCode'],
            is_whitelisted=data['whitelisted'],
            description=data['gameDescription'],
            max_players=data['maxPlayers'],
            game_version=data['gameVersion'],
            has_password=data['hasPassword'],
            server_url=data['serverUrl'],
            token=base64.b64encode(secrets.token_bytes(32)).decode('utf-8')
        )
        server.save()
        mods = [VSMod(_id=mod['id'], version=mod["version"], server=server) for mod in data['Mods']]
        for mod in mods: mod.save()
        return server


class VSServer(models.Model):
    objects = VSServerManager()

    vs_server_name = models.CharField(max_length=512)
    address = models.CharField(max_length=256)  # IP, URL probably acceptable so 256
    port = models.PositiveIntegerField()

    play_style_id = models.CharField(max_length=64)
    play_style_lang_code = models.CharField(max_length=64)
    is_whitelisted = models.BooleanField()
    description = models.TextField()
    max_players = models.SmallIntegerField()
    game_version = models.CharField(max_length=64)
    has_password = models.BooleanField()
    server_url = models.CharField(max_length=256, null=True)
    last_heartbeat = models.DateTimeField(auto_now_add=True)
    players = models.SmallIntegerField(default=0)

    token = models.CharField(max_length=64)

    def as_dict(self):
        return {
            "serverName": self.vs_server_name,
            "serverIP": f"{self.address}:{self.port}",
            "playstyle": {
                "id": self.play_style_id,
                "langCode": self.play_style_lang_code
            },
            "Mods": [{"id": mod._id, "version": mod.version} for mod in self.mods.all()],
            "maxPlayers": self.max_players,
            "gameVersion": self.game_version,
            "hasPassword": self.has_password,
            "whitelisted": self.is_whitelisted,
            "gameDescription": self.description,
        }

    def __repr__(self):
        return f"{self.vs_server_name}@{self.address}"

    def __str__(self):
        return f"{self.vs_server_name}@{self.address}"


class VSMod(models.Model):
    _id = models.CharField(max_length=64)
    version = models.CharField(max_length=64)
    server = models.ForeignKey(VSServer, on_delete=models.CASCADE, related_name='mods')

    def __str__(self):
        return f"{self._id} v{self.version}"



from django.utils import timezone

from .models import VSServer


def get_client_ip(request) -> str:
    """
    Extract ip address of request client, first checking for if a proxy is used
    """
    x_forwarded_for = request.META.get('HTTP_X_FORWARDED_FOR')
    if x_forwarded_for:
        ip = x_forwarded_for.split(',')[0]
    else:
        ip = request.META.get('REMOTE_ADDR')
    return ip

def vs_prune_old_servers(age_mins: int = 10) -> int:
    """
    Check database for entries that haven't been updated for over `age` minutes and delete them, assuming them to
    have gone offline.

    :returns: int as a count of server listings removed from the database.
    """
    counter = 0
    for server in VSServer.objects.all():
        if server.last_heartbeat < timezone.now() - timezone.timedelta(minutes=age_mins):
            server.delete()
            counter += 1
    return counter
